from typing import Literal, Union, Annotated
from pydantic import BaseModel, Field, TypeAdapter

try:
    from aiortc import RTCSessionDescription, RTCIceCandidate
except ImportError:
    RTCSessionDescription = None  # type: ignore
    RTCIceCandidate = None  # type: ignore


# Incoming Messages (from server)


class AnswerMessage(BaseModel):
    """WebRTC answer from server."""

    type: Literal["answer"]
    sdp: str


class IceCandidatePayload(BaseModel):
    """ICE candidate data."""

    candidate: str
    sdpMLineIndex: int
    sdpMid: str


class IceCandidateMessage(BaseModel):
    """ICE candidate message."""

    type: Literal["ice-candidate"]
    candidate: IceCandidatePayload


class SessionIdMessage(BaseModel):
    """Session initialization message from server."""

    type: Literal["session_id"]
    session_id: str
    server_port: int
    server_ip: str


# Discriminated union for incoming messages
IncomingMessage = Annotated[
    Union[AnswerMessage, IceCandidateMessage, SessionIdMessage],
    Field(discriminator="type"),
]

# Type adapter for parsing incoming messages
IncomingMessageAdapter = TypeAdapter(IncomingMessage)


# Outgoing Messages (to server)


class OfferMessage(BaseModel):
    """WebRTC offer to server."""

    type: Literal["offer"]
    sdp: str


class PromptMessage(BaseModel):
    """Update prompt message."""

    type: Literal["prompt"]
    prompt: str


class SwitchCameraMessage(BaseModel):
    """Switch camera/mirror message."""

    type: Literal["switch_camera"]
    rotateY: int


# Outgoing message union (no discriminator needed - we know what we're sending)
OutgoingMessage = Union[OfferMessage, IceCandidateMessage, PromptMessage, SwitchCameraMessage]


def parse_incoming_message(data: dict) -> IncomingMessage:
    """
    Parse incoming WebSocket message.

    Args:
        data: Message dictionary

    Returns:
        Parsed message instance

    Raises:
        ValidationError: If message format is invalid
    """
    return IncomingMessageAdapter.validate_python(data)


def message_to_json(message: OutgoingMessage) -> str:
    """
    Serialize outgoing message to JSON.

    Args:
        message: Message to serialize

    Returns:
        JSON string
    """
    return message.model_dump_json()



import pytest
from unittest.mock import AsyncMock, MagicMock, patch
from decart import DecartClient, models

try:
    from decart.realtime.client import RealtimeClient

    REALTIME_AVAILABLE = True
except ImportError:
    REALTIME_AVAILABLE = False

pytestmark = pytest.mark.skipif(
    not REALTIME_AVAILABLE,
    reason="Realtime API not available - install with: pip install decart[realtime]",
)


def test_realtime_client_available():
    """Test that realtime client is available when aiortc is installed"""
    assert REALTIME_AVAILABLE
    assert RealtimeClient is not None


def test_realtime_models_available():
    """Test that realtime models are available"""
    model = models.realtime("mirage")
    assert model.name == "mirage"
    assert model.fps == 25
    assert model.width == 1280
    assert model.height == 704

    model2 = models.realtime("lucy_v2v_720p_rt")
    assert model2.name == "lucy_v2v_720p_rt"


@pytest.mark.asyncio
async def test_realtime_client_creation_with_mock():
    """Test client creation with mocked WebRTC"""
    client = DecartClient(api_key="test-key")

    with patch("decart.realtime.client.WebRTCManager") as mock_manager_class:
        mock_manager = AsyncMock()
        mock_manager.connect = AsyncMock(return_value=True)
        mock_manager.is_connected = MagicMock(return_value=True)
        mock_manager.get_connection_state = MagicMock(return_value="connected")
        mock_manager_class.return_value = mock_manager

        mock_track = MagicMock()

        from decart.realtime.types import RealtimeConnectOptions
        from decart.types import ModelState, Prompt

        realtime_client = await RealtimeClient.connect(
            base_url=client.base_url,
            api_key=client.api_key,
            local_track=mock_track,
            options=RealtimeConnectOptions(
                model=models.realtime("mirage"),
                on_remote_stream=lambda t: None,
                initial_state=ModelState(prompt=Prompt(text="Test", enrich=True), mirror=False),
            ),
        )

        assert realtime_client is not None
        assert realtime_client.session_id
        assert realtime_client.is_connected()


@pytest.mark.asyncio
async def test_realtime_set_prompt_with_mock():
    """Test set_prompt with mocked WebRTC"""
    client = DecartClient(api_key="test-key")

    with patch("decart.realtime.client.WebRTCManager") as mock_manager_class:
        mock_manager = AsyncMock()
        mock_manager.connect = AsyncMock(return_value=True)
        mock_manager.send_message = AsyncMock()
        mock_manager_class.return_value = mock_manager

        mock_track = MagicMock()

        from decart.realtime.types import RealtimeConnectOptions

        realtime_client = await RealtimeClient.connect(
            base_url=client.base_url,
            api_key=client.api_key,
            local_track=mock_track,
            options=RealtimeConnectOptions(
                model=models.realtime("mirage"),
                on_remote_stream=lambda t: None,
            ),
        )

        await realtime_client.set_prompt("New prompt")

        mock_manager.send_message.assert_called_once()
        call_args = mock_manager.send_message.call_args[0][0]
        assert call_args.type == "prompt"
        assert call_args.prompt == "New prompt"


@pytest.mark.asyncio
async def test_realtime_set_mirror_with_mock():
    """Test set_mirror with mocked WebRTC"""
    client = DecartClient(api_key="test-key")

    with patch("decart.realtime.client.WebRTCManager") as mock_manager_class:
        mock_manager = AsyncMock()
        mock_manager.connect = AsyncMock(return_value=True)
        mock_manager.send_message = AsyncMock()
        mock_manager_class.return_value = mock_manager

        mock_track = MagicMock()

        from decart.realtime.types import RealtimeConnectOptions

        realtime_client = await RealtimeClient.connect(
            base_url=client.base_url,
            api_key=client.api_key,
            local_track=mock_track,
            options=RealtimeConnectOptions(
                model=models.realtime("mirage"),
                on_remote_stream=lambda t: None,
            ),
        )

        await realtime_client.set_mirror(True)

        mock_manager.send_message.assert_called_once()
        call_args = mock_manager.send_message.call_args[0][0]
        assert call_args.type == "switch_camera"
        assert call_args.rotateY == 2


@pytest.mark.asyncio
async def test_realtime_events():
    """Test event handling"""
    client = DecartClient(api_key="test-key")

    with patch("decart.realtime.client.WebRTCManager") as mock_manager_class:
        mock_manager = AsyncMock()
        mock_manager.connect = AsyncMock(return_value=True)
        mock_manager_class.return_value = mock_manager

        mock_track = MagicMock()

        from decart.realtime.types import RealtimeConnectOptions

        realtime_client = await RealtimeClient.connect(
            base_url=client.base_url,
            api_key=client.api_key,
            local_track=mock_track,
            options=RealtimeConnectOptions(
                model=models.realtime("mirage"),
                on_remote_stream=lambda t: None,
            ),
        )

        connection_states = []
        errors = []

        def on_connection_change(state):
            connection_states.append(state)

        def on_error(error):
            errors.append(error)

        realtime_client.on("connection_change", on_connection_change)
        realtime_client.on("error", on_error)

        realtime_client._emit_connection_change("connected")
        assert connection_states == ["connected"]

        from decart.errors import DecartSDKError

        test_error = DecartSDKError("Test error")
        realtime_client._emit_error(test_error)
        assert len(errors) == 1
        assert errors[0].message == "Test error"



import asyncio
import os
from decart import DecartClient, models


async def main() -> None:
    async with DecartClient(api_key=os.getenv("DECART_API_KEY", "your-api-key-here")) as client:
        print("Generating image from text...")
        result = await client.process(
            {
                "model": models.image("lucy-pro-t2i"),
                "prompt": "A futuristic cityscape at night with neon lights",
                "seed": 42,
                "orientation": "portrait",
            }
        )

        with open("output_t2i.png", "wb") as f:
            f.write(result)

        print("Image saved to output_t2i.png")

        print("\nTransforming image...")
        with open("output_t2i.png", "rb") as image_file:
            result = await client.process(
                {
                    "model": models.image("lucy-pro-i2i"),
                    "prompt": "Oil painting style with impressionist brushstrokes",
                    "data": image_file,
                    "enhance_prompt": True,
                }
            )

        with open("output_i2i.png", "wb") as f:
            f.write(result)

        print("Image saved to output_i2i.png")


if __name__ == "__main__":
    asyncio.run(main())



from typing import Literal, Optional
from pydantic import BaseModel, Field, ConfigDict
from .errors import ModelNotFoundError
from .types import FileInput


RealTimeModels = Literal["mirage", "lucy_v2v_720p_rt"]
VideoModels = Literal[
    "lucy-dev-i2v",
    "lucy-dev-v2v",
    "lucy-pro-t2v",
    "lucy-pro-i2v",
    "lucy-pro-v2v",
    "lucy-pro-flf2v",
]
ImageModels = Literal["lucy-pro-t2i", "lucy-pro-i2i"]
Model = Literal[RealTimeModels, VideoModels, ImageModels]


class DecartBaseModel(BaseModel):
    model_config = ConfigDict(arbitrary_types_allowed=True)


class ModelDefinition(DecartBaseModel):
    name: str
    url_path: str
    fps: int = Field(ge=1)
    width: int = Field(ge=1)
    height: int = Field(ge=1)
    input_schema: type[BaseModel]


class TextToVideoInput(BaseModel):
    prompt: str = Field(..., min_length=1)
    seed: Optional[int] = None
    resolution: Optional[str] = None
    orientation: Optional[str] = None


class ImageToVideoInput(DecartBaseModel):
    prompt: str = Field(..., min_length=1)
    data: FileInput
    seed: Optional[int] = None
    resolution: Optional[str] = None


class VideoToVideoInput(DecartBaseModel):
    prompt: str = Field(..., min_length=1)
    data: FileInput
    seed: Optional[int] = None
    resolution: Optional[str] = None
    enhance_prompt: Optional[bool] = None
    num_inference_steps: Optional[int] = None


class FirstLastFrameInput(DecartBaseModel):
    prompt: str = Field(..., min_length=1)
    start: FileInput
    end: FileInput
    seed: Optional[int] = None
    resolution: Optional[str] = None


class TextToImageInput(BaseModel):
    prompt: str = Field(..., min_length=1)
    seed: Optional[int] = None
    resolution: Optional[str] = None
    orientation: Optional[str] = None


class ImageToImageInput(DecartBaseModel):
    prompt: str = Field(..., min_length=1)
    data: FileInput
    seed: Optional[int] = None
    resolution: Optional[str] = None
    enhance_prompt: Optional[bool] = None


_MODELS = {
    "realtime": {
        "mirage": ModelDefinition(
            name="mirage",
            url_path="/v1/stream",
            fps=25,
            width=1280,
            height=704,
            input_schema=BaseModel,
        ),
        "lucy_v2v_720p_rt": ModelDefinition(
            name="lucy_v2v_720p_rt",
            url_path="/v1/stream",
            fps=25,
            width=1280,
            height=704,
            input_schema=BaseModel,
        ),
    },
    "video": {
        "lucy-dev-i2v": ModelDefinition(
            name="lucy-dev-i2v",
            url_path="/v1/generate/lucy-dev-i2v",
            fps=25,
            width=1280,
            height=704,
            input_schema=ImageToVideoInput,
        ),
        "lucy-dev-v2v": ModelDefinition(
            name="lucy-dev-v2v",
            url_path="/v1/generate/lucy-dev-v2v",
            fps=25,
            width=1280,
            height=704,
            input_schema=VideoToVideoInput,
        ),
        "lucy-pro-t2v": ModelDefinition(
            name="lucy-pro-t2v",
            url_path="/v1/generate/lucy-pro-t2v",
            fps=25,
            width=1280,
            height=704,
            input_schema=TextToVideoInput,
        ),
        "lucy-pro-i2v": ModelDefinition(
            name="lucy-pro-i2v",
            url_path="/v1/generate/lucy-pro-i2v",
            fps=25,
            width=1280,
            height=704,
            input_schema=ImageToVideoInput,
        ),
        "lucy-pro-v2v": ModelDefinition(
            name="lucy-pro-v2v",
            url_path="/v1/generate/lucy-pro-v2v",
            fps=25,
            width=1280,
            height=704,
            input_schema=VideoToVideoInput,
        ),
        "lucy-pro-flf2v": ModelDefinition(
            name="lucy-pro-flf2v",
            url_path="/v1/generate/lucy-pro-flf2v",
            fps=25,
            width=1280,
            height=704,
            input_schema=FirstLastFrameInput,
        ),
    },
    "image": {
        "lucy-pro-t2i": ModelDefinition(
            name="lucy-pro-t2i",
            url_path="/v1/generate/lucy-pro-t2i",
            fps=25,
            width=1280,
            height=704,
            input_schema=TextToImageInput,
        ),
        "lucy-pro-i2i": ModelDefinition(
            name="lucy-pro-i2i",
            url_path="/v1/generate/lucy-pro-i2i",
            fps=25,
            width=1280,
            height=704,
            input_schema=ImageToImageInput,
        ),
    },
}


class Models:
    @staticmethod
    def realtime(model: RealTimeModels) -> ModelDefinition:
        try:
            return _MODELS["realtime"][model]
        except KeyError:
            raise ModelNotFoundError(model)

    @staticmethod
    def video(model: VideoModels) -> ModelDefinition:
        try:
            return _MODELS["video"][model]
        except KeyError:
            raise ModelNotFoundError(model)

    @staticmethod
    def image(model: ImageModels) -> ModelDefinition:
        try:
            return _MODELS["image"][model]
        except KeyError:
            raise ModelNotFoundError(model)


models = Models()



import pytest
import asyncio
from unittest.mock import AsyncMock, patch, MagicMock
from decart import DecartClient, models, DecartSDKError


@pytest.mark.asyncio
async def test_process_text_to_video() -> None:
    client = DecartClient(api_key="test-key")

    with patch("aiohttp.ClientSession") as mock_session_cls:
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.read = AsyncMock(return_value=b"fake video data")

        mock_session = MagicMock()
        mock_session.__aenter__ = AsyncMock(return_value=mock_session)
        mock_session.__aexit__ = AsyncMock(return_value=None)
        mock_session.post = MagicMock()
        mock_session.post.return_value.__aenter__ = AsyncMock(return_value=mock_response)
        mock_session.post.return_value.__aexit__ = AsyncMock(return_value=None)

        mock_session_cls.return_value = mock_session

        result = await client.process(
            {
                "model": models.video("lucy-pro-t2v"),
                "prompt": "A cat walking",
            }
        )

        assert result == b"fake video data"


@pytest.mark.asyncio
async def test_process_missing_model() -> None:
    client = DecartClient(api_key="test-key")

    with pytest.raises(DecartSDKError):
        await client.process(
            {
                "prompt": "A cat walking",
            }
        )


@pytest.mark.asyncio
async def test_process_missing_required_field() -> None:
    client = DecartClient(api_key="test-key")

    with pytest.raises(DecartSDKError):
        await client.process(
            {
                "model": models.video("lucy-pro-i2v"),
            }
        )


@pytest.mark.asyncio
async def test_process_video_to_video() -> None:
    client = DecartClient(api_key="test-key")

    with patch("aiohttp.ClientSession") as mock_session_cls:
        mock_response = MagicMock()
        mock_response.ok = True
        mock_response.read = AsyncMock(return_value=b"fake video data")

        mock_session = MagicMock()
        mock_session.__aenter__ = AsyncMock(return_value=mock_session)
        mock_session.__aexit__ = AsyncMock(return_value=None)
        mock_session.post = MagicMock()
        mock_session.post.return_value.__aenter__ = AsyncMock(return_value=mock_response)
        mock_session.post.return_value.__aexit__ = AsyncMock(return_value=None)

        mock_session_cls.return_value = mock_session

        result = await client.process(
            {
                "model": models.video("lucy-pro-v2v"),
                "prompt": "Anime style",
                "data": b"fake input video",
                "enhance_prompt": True,
            }
        )

        assert result == b"fake video data"


@pytest.mark.asyncio
async def test_process_with_cancellation() -> None:
    client = DecartClient(api_key="test-key")
    cancel_token = asyncio.Event()

    cancel_token.set()

    with pytest.raises(asyncio.CancelledError):
        await client.process(
            {
                "model": models.video("lucy-pro-t2v"),
                "prompt": "A video that will be cancelled",
                "cancel_token": cancel_token,
            }
        )



from typing import Union, Optional, Protocol, runtime_checkable
from pathlib import Path
from pydantic import BaseModel, Field


@runtime_checkable
class HasRead(Protocol):
    """Protocol for file-like objects with a read method."""

    def read(self) -> Union[bytes, str]: ...


FileInput = Union[HasRead, bytes, str, Path]


class Prompt(BaseModel):
    text: str = Field(..., min_length=1)
    enrich: bool = Field(default=True)


class ModelState(BaseModel):
    prompt: Optional[Prompt] = None
    mirror: bool = Field(default=False)



import asyncio
import os
from pathlib import Path
from decart import DecartClient, models

try:
    from aiortc.contrib.media import MediaPlayer, MediaRecorder
except ImportError:
    print("aiortc is required for this example.")
    print("Install with: pip install decart[realtime]")
    exit(1)


async def main():
    api_key = os.getenv("DECART_API_KEY")
    if not api_key:
        print("Error: DECART_API_KEY environment variable not set")
        print("Usage: DECART_API_KEY=your-key python realtime_file.py <video_file>")
        return

    import sys

    if len(sys.argv) < 2:
        print("Usage: python realtime_file.py <video_file>")
        print("Example: python realtime_file.py output_t2v.mp4")
        return

    video_file = sys.argv[1]
    if not os.path.exists(video_file):
        print(f"Error: Video file not found: {video_file}")
        return

    print(f"Loading video file: {video_file}")
    player = MediaPlayer(video_file)

    if not player.video:
        print("Error: No video stream found in file")
        return

    try:
        from decart.realtime.client import RealtimeClient
    except ImportError:
        print("Error: Realtime API not available")
        print("Install with: pip install decart[realtime]")
        return

    print("Creating Decart client...")
    async with DecartClient(api_key=api_key) as client:
        model = models.realtime("mirage")
        print(f"Using model: {model.name}")

        frame_count = 0
        recorder = None
        input_name = Path(video_file).stem
        output_file = Path(f"output_realtime_{input_name}.mp4")

        def on_remote_stream(track):
            nonlocal frame_count, recorder
            frame_count += 1
            if frame_count % 25 == 0:
                print(f"📹 Processed {frame_count} frames...")

            if recorder is None:
                print(f"💾 Recording to {output_file}")
                recorder = MediaRecorder(str(output_file))
                recorder.addTrack(track)
                asyncio.create_task(recorder.start())

        def on_connection_change(state):
            print(f"🔄 Connection state: {state}")

        def on_error(error):
            print(f"❌ Error: {error.__class__.__name__} - {error.message}")

        print("\nConnecting to Realtime API...")
        try:
            from decart.realtime.client import RealtimeClient
            from decart.realtime.types import RealtimeConnectOptions
            from decart.types import ModelState, Prompt

            realtime_client = await RealtimeClient.connect(
                base_url=client.base_url,
                api_key=client.api_key,
                local_track=player.video,
                options=RealtimeConnectOptions(
                    model=model,
                    on_remote_stream=on_remote_stream,
                    initial_state=ModelState(
                        prompt=Prompt(text="Lego World", enrich=True), mirror=False
                    ),
                ),
            )

            realtime_client.on("connection_change", on_connection_change)
            realtime_client.on("error", on_error)

            print("✓ Connected!")
            print(f"Session ID: {realtime_client.session_id}")
            print("Processing video... (Ctrl+C to stop)")

            try:
                while True:
                    await asyncio.sleep(1)
            except KeyboardInterrupt:
                print(f"\n\n✓ Processed {frame_count} frames total")
            finally:
                if recorder:
                    try:
                        print(f"💾 Saving video to {output_file}...")
                        await asyncio.sleep(0.5)
                        await recorder.stop()
                        print(f"✓ Video saved to {output_file}")
                    except Exception as e:
                        print(f"⚠️ Warning: Could not save video cleanly: {e}")
                        print("   Video file may be incomplete or corrupted")

        except Exception as e:
            print(f"\n❌ Connection failed: {e}")
            import traceback

            traceback.print_exc()
        finally:
            if "realtime_client" in locals():
                print("\nDisconnecting...")
                await realtime_client.disconnect()
                print("✓ Disconnected")


if __name__ == "__main__":
    asyncio.run(main())



import asyncio
import json
import logging
from typing import Optional, Callable
import aiohttp
from aiortc import (
    RTCPeerConnection,
    RTCSessionDescription,
    RTCIceCandidate,
    RTCConfiguration,
    RTCIceServer,
    MediaStreamTrack,
)

from ..errors import WebRTCError
from .messages import (
    parse_incoming_message,
    message_to_json,
    OfferMessage,
    IceCandidateMessage,
    IceCandidatePayload,
    OutgoingMessage,
)
from .types import ConnectionState

logger = logging.getLogger(__name__)


class WebRTCConnection:
    def __init__(
        self,
        on_remote_stream: Optional[Callable[[MediaStreamTrack], None]] = None,
        on_state_change: Optional[Callable[[ConnectionState], None]] = None,
        on_error: Optional[Callable[[Exception], None]] = None,
        customize_offer: Optional[Callable] = None,
    ):

        self._pc: Optional[RTCPeerConnection] = None
        self._ws: Optional[aiohttp.ClientWebSocketResponse] = None
        self._session: Optional[aiohttp.ClientSession] = None
        self._state: ConnectionState = "disconnected"
        self._on_remote_stream = on_remote_stream
        self._on_state_change = on_state_change
        self._on_error = on_error
        self._customize_offer = customize_offer
        self._ws_task: Optional[asyncio.Task] = None
        self._ice_candidates_queue: list[RTCIceCandidate] = []

    async def connect(self, url: str, local_track: MediaStreamTrack, timeout: float = 30) -> None:
        try:
            await self._set_state("connecting")

            ws_url = url.replace("https://", "wss://").replace("http://", "ws://")

            self._session = aiohttp.ClientSession()
            self._ws = await self._session.ws_connect(ws_url)

            self._ws_task = asyncio.create_task(self._receive_messages())

            await self._setup_peer_connection(local_track)

            await self._create_and_send_offer()

            deadline = asyncio.get_event_loop().time() + timeout
            while asyncio.get_event_loop().time() < deadline:
                if self._state == "connected":
                    return
                await asyncio.sleep(0.1)

            raise TimeoutError("Connection timeout")

        except Exception as e:
            logger.error(f"Connection failed: {e}")
            await self._set_state("disconnected")
            if self._on_error:
                self._on_error(e)
            raise WebRTCError(str(e), cause=e)

    async def _setup_peer_connection(self, local_track: MediaStreamTrack) -> None:
        config = RTCConfiguration(iceServers=[RTCIceServer(urls=["stun:stun.l.google.com:19302"])])

        self._pc = RTCPeerConnection(configuration=config)

        @self._pc.on("track")
        def on_track(track: MediaStreamTrack):
            logger.debug(f"Received remote track: {track.kind}")
            if self._on_remote_stream:
                self._on_remote_stream(track)

        @self._pc.on("icecandidate")
        async def on_ice_candidate(candidate: RTCIceCandidate):
            if candidate:
                logger.debug(f"Local ICE candidate: {candidate.candidate}")
                await self._send_message(
                    IceCandidateMessage(
                        type="ice-candidate",
                        candidate=IceCandidatePayload(
                            candidate=candidate.candidate,
                            sdpMLineIndex=candidate.sdpMLineIndex or 0,
                            sdpMid=candidate.sdpMid or "",
                        ),
                    )
                )

        @self._pc.on("connectionstatechange")
        async def on_connection_state_change():
            logger.debug(f"Peer connection state: {self._pc.connectionState}")
            if self._pc.connectionState == "connected":
                await self._set_state("connected")
            elif self._pc.connectionState in ["failed", "closed"]:
                await self._set_state("disconnected")

        @self._pc.on("iceconnectionstatechange")
        async def on_ice_connection_state_change():
            logger.debug(f"ICE connection state: {self._pc.iceConnectionState}")

        self._pc.addTrack(local_track)
        logger.debug("Added local track to peer connection")

    async def _create_and_send_offer(self) -> None:
        logger.debug("Creating offer...")

        offer = await self._pc.createOffer()
        logger.debug(f"Offer SDP:\n{offer.sdp}")

        if self._customize_offer:
            await self._customize_offer(offer)

        await self._pc.setLocalDescription(offer)
        logger.debug("Set local description (offer)")

        await self._send_message(OfferMessage(type="offer", sdp=self._pc.localDescription.sdp))

    async def _receive_messages(self) -> None:
        try:
            async for msg in self._ws:
                if msg.type == aiohttp.WSMsgType.TEXT:
                    try:
                        data = json.loads(msg.data)
                        logger.debug(f"Received {data.get('type', 'unknown')} message")
                        logger.debug(f"Message content: {msg.data}")
                        await self._handle_message(data)
                    except Exception as e:
                        logger.error(f"Error handling message: {e}")
                elif msg.type == aiohttp.WSMsgType.ERROR:
                    logger.error(f"WebSocket error: {self._ws.exception()}")
                    break
        except Exception as e:
            logger.error(f"WebSocket receive error: {e}")
            if self._on_error:
                self._on_error(e)

    async def _handle_message(self, data: dict) -> None:
        try:
            message = parse_incoming_message(data)
        except Exception as e:
            logger.warning(f"Failed to parse message: {e}")
            return

        if message.type == "answer":
            await self._handle_answer(message.sdp)
        elif message.type == "ice-candidate":
            await self._handle_ice_candidate(message.candidate)
        elif message.type == "session_id":
            logger.debug(f"Session ID: {message.session_id}")

    async def _handle_answer(self, sdp: str) -> None:
        logger.debug("Received answer from server")
        logger.debug(f"Answer SDP:\n{sdp}")

        answer = RTCSessionDescription(sdp=sdp, type="answer")
        await self._pc.setRemoteDescription(answer)
        logger.debug("Set remote description (answer)")

        if self._ice_candidates_queue:
            logger.debug(f"Adding {len(self._ice_candidates_queue)} queued ICE candidates")
            for candidate in self._ice_candidates_queue:
                await self._pc.addIceCandidate(candidate)
            self._ice_candidates_queue.clear()

    async def _handle_ice_candidate(self, candidate_data: IceCandidatePayload) -> None:
        logger.debug(f"Remote ICE candidate: {candidate_data.candidate}")

        candidate = RTCIceCandidate(
            candidate=candidate_data.candidate,
            sdpMLineIndex=candidate_data.sdpMLineIndex,
            sdpMid=candidate_data.sdpMid,
        )

        if self._pc.remoteDescription:
            logger.debug("Adding ICE candidate to peer connection")
            await self._pc.addIceCandidate(candidate)
        else:
            logger.debug("Queuing ICE candidate (no remote description yet)")
            self._ice_candidates_queue.append(candidate)

    async def _send_message(self, message: OutgoingMessage) -> None:
        if not self._ws or self._ws.closed:
            raise RuntimeError("WebSocket not connected")

        msg_json = message_to_json(message)
        logger.debug(f"Sending {message.type} message")
        logger.debug(f"Message content: {msg_json}")
        await self._ws.send_str(msg_json)

    async def _set_state(self, state: ConnectionState) -> None:
        if self._state != state:
            self._state = state
            logger.debug(f"Connection state changed to: {state}")
            if self._on_state_change:
                self._on_state_change(state)

    async def send(self, message: OutgoingMessage) -> None:
        await self._send_message(message)

    @property
    def state(self) -> ConnectionState:
        return self._state

    async def cleanup(self) -> None:
        if self._ws_task:
            self._ws_task.cancel()
            try:
                await self._ws_task
            except asyncio.CancelledError:
                pass

        if self._pc:
            await self._pc.close()

        if self._ws and not self._ws.closed:
            await self._ws.close()

        if self._session and not self._session.closed:
            await self._session.close()

        await self._set_state("disconnected")



from typing import Any, Optional


class DecartSDKError(Exception):
    """Base exception for all Decart SDK errors."""

    def __init__(
        self,
        message: str,
        data: Optional[dict[str, Any]] = None,
        cause: Optional[Exception] = None,
    ) -> None:
        self.message = message
        self.data = data or {}
        self.cause = cause
        super().__init__(message)

    def __str__(self) -> str:
        return self.message

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({self.message!r})"


class InvalidAPIKeyError(DecartSDKError):
    """Raised when API key is invalid or missing."""

    def __init__(self) -> None:
        super().__init__("API key is required and must be a non-empty string")


class InvalidBaseURLError(DecartSDKError):
    """Raised when base URL is invalid."""

    def __init__(self, url: Optional[str] = None) -> None:
        message = f"Invalid base URL: {url}" if url else "Invalid base URL"
        super().__init__(message)


class WebRTCError(DecartSDKError):
    """Raised when WebRTC connection fails."""

    def __init__(self, message: str = "WebRTC error", cause: Optional[Exception] = None) -> None:
        super().__init__(message, cause=cause)


class InvalidInputError(DecartSDKError):
    """Raised when input validation fails."""

    pass


class ModelNotFoundError(DecartSDKError):
    """Raised when model is not found."""

    def __init__(self, model: str) -> None:
        super().__init__(f"Model {model} not found")
        self.model = model


class ProcessingError(DecartSDKError):
    """Raised when processing fails."""

    pass



from typing import Literal, Callable, Optional
from dataclasses import dataclass
from ..models import ModelDefinition
from ..types import ModelState

try:
    from aiortc import MediaStreamTrack
except ImportError:
    MediaStreamTrack = None  # type: ignore


ConnectionState = Literal["connecting", "connected", "disconnected"]


@dataclass
class RealtimeConnectOptions:
    model: ModelDefinition
    on_remote_stream: Callable[[MediaStreamTrack], None]
    initial_state: Optional[ModelState] = None
    customize_offer: Optional[Callable] = None



import asyncio
import os
from decart import DecartClient, models


async def main() -> None:
    async with DecartClient(api_key=os.getenv("DECART_API_KEY", "your-api-key-here")) as client:
        print("Transforming video from URL...")
        result = await client.process(
            {
                "model": models.video("lucy-pro-v2v"),
                "prompt": "Watercolor painting style",
                "data": "https://docs.platform.decart.ai/assets/example-video.mp4",
            }
        )

        with open("output_url.mp4", "wb") as f:
            f.write(result)

        print("Video saved to output_url.mp4")


if __name__ == "__main__":
    asyncio.run(main())



import aiohttp
import aiofiles
import asyncio
from pathlib import Path
from typing import Any, Optional
from ..types import FileInput
from ..models import ModelDefinition
from ..errors import InvalidInputError, ProcessingError


async def file_input_to_bytes(
    input_data: FileInput, session: aiohttp.ClientSession
) -> tuple[bytes, str]:
    """Convert various file input types to bytes asynchronously.

    Args:
        input_data: The file input (bytes, Path, str, or file-like object)
        session: Reusable aiohttp session for URL fetching

    Returns:
        Tuple of (content bytes, content type)

    Raises:
        InvalidInputError: If input is invalid or processing fails
    """

    if isinstance(input_data, bytes):
        return input_data, "application/octet-stream"

    if isinstance(input_data, Path):
        # Async file reading with aiofiles
        try:
            async with aiofiles.open(input_data, mode="rb") as f:
                content = await f.read()
            return content, "application/octet-stream"
        except FileNotFoundError:
            raise InvalidInputError(f"File not found: {input_data}")
        except Exception as e:
            raise InvalidInputError(f"Failed to read file {input_data}: {str(e)}")

    if isinstance(input_data, str):
        # Check if it's a file path
        path = Path(input_data)
        if path.exists():
            try:
                async with aiofiles.open(path, mode="rb") as f:
                    content = await f.read()
                return content, "application/octet-stream"
            except Exception as e:
                raise InvalidInputError(f"Failed to read file {input_data}: {str(e)}")

        # Otherwise treat as URL
        if not input_data.startswith(("http://", "https://")):
            raise InvalidInputError(
                f"Input must be a URL (http:// or https://) or existing file path: {input_data}"
            )

        # Use the provided session instead of creating a new one
        async with session.get(input_data) as response:
            if not response.ok:
                raise InvalidInputError(f"Failed to fetch file from URL: {response.status}")
            content = await response.read()
            content_type = response.headers.get("Content-Type", "application/octet-stream")
            return content, content_type

    from ..types import HasRead

    if isinstance(input_data, HasRead):
        # Sync file-like objects (for backwards compatibility)
        content = await asyncio.to_thread(input_data.read)
        if isinstance(content, str):
            content = content.encode()
        return content, "application/octet-stream"

    raise InvalidInputError(f"Invalid file input type: {type(input_data)}")


async def send_request(
    session: aiohttp.ClientSession,
    base_url: str,
    api_key: str,
    model: ModelDefinition,
    inputs: dict[str, Any],
    cancel_token: Optional[asyncio.Event] = None,
) -> bytes:
    form_data = aiohttp.FormData()

    for key, value in inputs.items():
        if value is not None:
            if key in ("data", "start", "end"):
                content, content_type = await file_input_to_bytes(value, session)
                form_data.add_field(key, content, content_type=content_type)
            else:
                form_data.add_field(key, str(value))

    endpoint = f"{base_url}{model.url_path}"

    async def make_request() -> bytes:
        async with session.post(
            endpoint,
            headers={"X-API-KEY": api_key},
            data=form_data,
        ) as response:
            if not response.ok:
                error_text = await response.text()
                raise ProcessingError(f"Processing failed: {response.status} - {error_text}")
            return await response.read()

    if cancel_token:
        request_task = asyncio.create_task(make_request())
        cancel_task = asyncio.create_task(cancel_token.wait())

        done, pending = await asyncio.wait(
            [request_task, cancel_task], return_when=asyncio.FIRST_COMPLETED
        )

        for task in pending:
            task.cancel()

        if cancel_task in done:
            try:
                await request_task
            except asyncio.CancelledError:
                pass
            raise asyncio.CancelledError("Request cancelled by user")

        return await request_task

    return await make_request()



from typing import Any, Optional
import aiohttp
from pydantic import ValidationError
from .errors import InvalidAPIKeyError, InvalidBaseURLError, InvalidInputError
from .models import ModelDefinition
from .process.request import send_request

try:
    from .realtime.client import RealtimeClient

    REALTIME_AVAILABLE = True
except ImportError:
    REALTIME_AVAILABLE = False
    RealtimeClient = None  # type: ignore


class DecartClient:
    """
    Decart API client for video and image generation/transformation.

    Args:
        api_key: Your Decart API key
        base_url: API base URL (defaults to production)

    Example:
        ```python
        client = DecartClient(api_key="your-key")
        result = await client.process({
            "model": models.video("lucy-pro-t2v"),
            "prompt": "A serene lake at sunset",
        })
        ```
    """

    def __init__(self, api_key: str, base_url: str = "https://api.decart.ai") -> None:
        if not api_key or not api_key.strip():
            raise InvalidAPIKeyError()

        if not base_url.startswith(("http://", "https://")):
            raise InvalidBaseURLError(base_url)

        self.api_key = api_key
        self.base_url = base_url
        self._session: Optional[aiohttp.ClientSession] = None

    async def _get_session(self) -> aiohttp.ClientSession:
        """Get or create the aiohttp session."""
        if self._session is None or self._session.closed:
            timeout = aiohttp.ClientTimeout(total=300)
            self._session = aiohttp.ClientSession(timeout=timeout)
        return self._session

    async def close(self) -> None:
        """Close the HTTP session and cleanup resources."""
        if self._session and not self._session.closed:
            await self._session.close()

    async def __aenter__(self):
        """Async context manager entry."""
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        """Async context manager exit."""
        await self.close()

    async def process(self, options: dict[str, Any]) -> bytes:
        """
        Process video or image generation/transformation.

        Args:
            options: Processing options including model and inputs

        Returns:
            Generated/transformed media as bytes

        Raises:
            InvalidInputError: If inputs are invalid
            ProcessingError: If processing fails
        """
        if "model" not in options:
            raise InvalidInputError("model is required")

        model: ModelDefinition = options["model"]
        cancel_token = options.get("cancel_token")

        inputs = {k: v for k, v in options.items() if k not in ("model", "cancel_token")}

        # File fields that need special handling (not validated by Pydantic)
        FILE_FIELDS = {"data", "start", "end"}

        # Separate file inputs from regular inputs
        file_inputs = {k: v for k, v in inputs.items() if k in FILE_FIELDS}
        non_file_inputs = {k: v for k, v in inputs.items() if k not in FILE_FIELDS}

        # Validate non-file inputs and create placeholder for file fields
        validation_inputs = {
            **non_file_inputs,
            **{k: b"" for k in file_inputs.keys()},  # Placeholder bytes for validation
        }

        try:
            validated_inputs = model.input_schema(**validation_inputs)
        except ValidationError as e:
            raise InvalidInputError(f"Invalid inputs for {model.name}: {str(e)}") from e

        # Build final inputs: validated non-file inputs + original file inputs
        processed_inputs = {
            **validated_inputs.model_dump(exclude_none=True),
            **file_inputs,  # Override placeholders with actual file data
        }

        session = await self._get_session()
        response = await send_request(
            session=session,
            base_url=self.base_url,
            api_key=self.api_key,
            model=model,
            inputs=processed_inputs,
            cancel_token=cancel_token,
        )

        return response



import pytest
from decart import DecartClient, InvalidAPIKeyError, InvalidBaseURLError


def test_create_client_success() -> None:
    client = DecartClient(api_key="test-key")
    assert client is not None
    assert client.process is not None


def test_create_client_invalid_api_key() -> None:
    with pytest.raises(InvalidAPIKeyError):
        DecartClient(api_key="")


def test_create_client_invalid_base_url() -> None:
    with pytest.raises(InvalidBaseURLError):
        DecartClient(api_key="test-key", base_url="invalid-url")


def test_create_client_custom_base_url() -> None:
    client = DecartClient(api_key="test-key", base_url="https://custom.decart.ai")
    assert client is not None
    assert client.base_url == "https://custom.decart.ai"



import asyncio
import os
from decart import DecartClient, models


async def main() -> None:
    async with DecartClient(api_key=os.getenv("DECART_API_KEY", "your-api-key-here")) as client:
        print("Generating video from text...")
        result = await client.process(
            {
                "model": models.video("lucy-pro-t2v"),
                "prompt": "A serene lake at sunset with mountains in the background",
                "seed": 42,
            }
        )

        with open("output_t2v.mp4", "wb") as f:
            f.write(result)

        print("Video saved to output_t2v.mp4")

        print("Transforming video...")
        with open("output_t2v.mp4", "rb") as video_file:
            result = await client.process(
                {
                    "model": models.video("lucy-pro-v2v"),
                    "prompt": "Anime style with vibrant colors",
                    "data": video_file,
                    "enhance_prompt": True,
                    "num_inference_steps": 50,
                }
            )

        with open("output_v2v.mp4", "wb") as f:
            f.write(result)

        print("Video saved to output_v2v.mp4")


if __name__ == "__main__":
    asyncio.run(main())



from .client import DecartClient
from .errors import (
    DecartSDKError,
    InvalidAPIKeyError,
    InvalidBaseURLError,
    InvalidInputError,
    ModelNotFoundError,
    ProcessingError,
    WebRTCError,
)
from .models import models, ModelDefinition
from .types import FileInput, ModelState, Prompt

try:
    from .realtime import (
        RealtimeClient,
        RealtimeConnectOptions,
        ConnectionState,
    )

    REALTIME_AVAILABLE = True
except ImportError:
    REALTIME_AVAILABLE = False
    RealtimeClient = None  # type: ignore
    RealtimeConnectOptions = None  # type: ignore
    ConnectionState = None  # type: ignore

__version__ = "0.0.1"

__all__ = [
    "DecartClient",
    "DecartSDKError",
    "InvalidAPIKeyError",
    "InvalidBaseURLError",
    "InvalidInputError",
    "ModelNotFoundError",
    "ProcessingError",
    "WebRTCError",
    "models",
    "ModelDefinition",
    "FileInput",
    "ModelState",
    "Prompt",
]

if REALTIME_AVAILABLE:
    __all__.extend(
        [
            "RealtimeClient",
            "RealtimeConnectOptions",
            "ConnectionState",
        ]
    )



import asyncio
import logging
import os
import numpy as np
from pathlib import Path
from decart import DecartClient, models

logging.basicConfig(
    level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s"
)

try:
    from aiortc import VideoStreamTrack
    from aiortc.contrib.media import MediaRecorder
    from av import VideoFrame
except ImportError:
    print("aiortc is required for this example.")
    print("Install with: pip install decart[realtime]")
    exit(1)


class SyntheticVideoTrack(VideoStreamTrack):
    """
    Generates synthetic video frames for testing.
    Creates colored frames that change over time.
    """

    def __init__(self):
        super().__init__()
        self.counter = 0

    async def recv(self):
        pts, time_base = await self.next_timestamp()

        colors = [
            (255, 0, 0),  # Red
            (0, 255, 0),  # Green
            (0, 0, 255),  # Blue
            (255, 255, 0),  # Yellow
        ]

        color_index = (self.counter // 25) % len(colors)
        color = colors[color_index]

        img = np.zeros((704, 1280, 3), dtype=np.uint8)
        img[:] = color

        frame = VideoFrame.from_ndarray(img, format="rgb24")
        frame.pts = pts
        frame.time_base = time_base

        self.counter += 1

        return frame


async def main():
    api_key = os.getenv("DECART_API_KEY")
    if not api_key:
        print("Error: DECART_API_KEY environment variable not set")
        print("Usage: DECART_API_KEY=your-key python realtime_synthetic.py")
        return

    try:
        from decart.realtime.client import RealtimeClient
    except ImportError:
        print("Error: Realtime API not available")
        print("Install with: pip install decart[realtime]")
        return

    print("Creating Decart client...")
    async with DecartClient(api_key=api_key) as client:
        print("Creating synthetic video track...")
        video_track = SyntheticVideoTrack()

        model = models.realtime("mirage")
        print(f"Using model: {model.name}")
        print(f"Model config - FPS: {model.fps}, Size: {model.width}x{model.height}")

        frame_count = 0
        recorder = None
        output_file = Path("output_realtime_synthetic.mp4")

        def on_remote_stream(track):
            nonlocal frame_count, recorder
            frame_count += 1
            print(f"📹 Received remote stream frame #{frame_count}")

            if recorder is None:
                print(f"💾 Recording to {output_file}")
                recorder = MediaRecorder(str(output_file))
                recorder.addTrack(track)
                asyncio.create_task(recorder.start())

        def on_connection_change(state):
            print(f"🔄 Connection state: {state}")

        def on_error(error):
            print(f"❌ Error: {error.__class__.__name__} - {error.message}")

        print("\nConnecting to Realtime API...")
        try:
            from decart.realtime.client import RealtimeClient
            from decart.realtime.types import RealtimeConnectOptions
            from decart.types import ModelState, Prompt

            realtime_client = await RealtimeClient.connect(
                base_url=client.base_url,
                api_key=client.api_key,
                local_track=video_track,
                options=RealtimeConnectOptions(
                    model=model,
                    on_remote_stream=on_remote_stream,
                    initial_state=ModelState(
                        prompt=Prompt(text="Anime style", enrich=True), mirror=False
                    ),
                ),
            )

            realtime_client.on("connection_change", on_connection_change)
            realtime_client.on("error", on_error)

            print("✓ Connected!")
            print(f"Session ID: {realtime_client.session_id}")
            print("Processing video for 10 seconds...")

            try:
                await asyncio.sleep(5)

                print("\n🎨 Changing style to 'Cyberpunk city'...")
                await realtime_client.set_prompt("Cyberpunk city")

                await asyncio.sleep(5)

                print(f"\n✓ Processed {frame_count} frames total")
            finally:
                if recorder:
                    try:
                        print(f"💾 Saving video to {output_file}...")
                        await asyncio.sleep(0.5)
                        await recorder.stop()
                        print(f"✓ Video saved to {output_file}")
                    except Exception as e:
                        print(f"⚠️ Warning: Could not save video cleanly: {e}")
                        print("   Video file may be incomplete or corrupted")

        except Exception as e:
            print(f"\n❌ Connection failed: {e}")
            import traceback

            traceback.print_exc()
        finally:
            if "realtime_client" in locals():
                print("\nDisconnecting...")
                await realtime_client.disconnect()
                print("✓ Disconnected")


if __name__ == "__main__":
    asyncio.run(main())



from typing import Callable
import logging
import uuid
from aiortc import MediaStreamTrack

from .webrtc_manager import WebRTCManager, WebRTCConfiguration
from .messages import PromptMessage, SwitchCameraMessage
from .types import ConnectionState, RealtimeConnectOptions
from ..errors import DecartSDKError, InvalidInputError, WebRTCError

logger = logging.getLogger(__name__)


class RealtimeClient:
    def __init__(self, manager: WebRTCManager, session_id: str):
        self._manager = manager
        self.session_id = session_id
        self._connection_callbacks: list[Callable[[ConnectionState], None]] = []
        self._error_callbacks: list[Callable[[DecartSDKError], None]] = []

    @classmethod
    async def connect(
        cls,
        base_url: str,
        api_key: str,
        local_track: MediaStreamTrack,
        options: RealtimeConnectOptions,
    ) -> "RealtimeClient":
        session_id = str(uuid.uuid4())
        ws_url = f"{base_url}{options.model.url_path}"
        ws_url += f"?api_key={api_key}&model={options.model.name}"

        config = WebRTCConfiguration(
            webrtc_url=ws_url,
            api_key=api_key,
            session_id=session_id,
            fps=options.model.fps,
            on_remote_stream=options.on_remote_stream,
            on_connection_state_change=None,
            on_error=None,
            initial_state=options.initial_state,
            customize_offer=options.customize_offer,
        )

        manager = WebRTCManager(config)
        client = cls(manager=manager, session_id=session_id)

        config.on_connection_state_change = client._emit_connection_change
        config.on_error = lambda error: client._emit_error(WebRTCError(str(error), cause=error))

        try:
            await manager.connect(local_track)

            if options.initial_state:
                if options.initial_state.prompt:
                    await client.set_prompt(
                        options.initial_state.prompt.text,
                        enrich=options.initial_state.prompt.enrich,
                    )
                if options.initial_state.mirror is not None:
                    await client.set_mirror(options.initial_state.mirror)
        except Exception as e:
            raise WebRTCError(str(e), cause=e)

        return client

    def _emit_connection_change(self, state: ConnectionState) -> None:
        for callback in self._connection_callbacks:
            try:
                callback(state)
            except Exception as e:
                logger.exception(f"Error in connection_change callback: {e}")

    def _emit_error(self, error: DecartSDKError) -> None:
        for callback in self._error_callbacks:
            try:
                callback(error)
            except Exception as e:
                logger.exception(f"Error in error callback: {e}")

    async def set_prompt(self, prompt: str, enrich: bool = True) -> None:
        if not prompt or not prompt.strip():
            raise InvalidInputError("Prompt cannot be empty")
        await self._manager.send_message(PromptMessage(type="prompt", prompt=prompt))

    async def set_mirror(self, enabled: bool) -> None:
        rotate_y = 2 if enabled else 0
        await self._manager.send_message(
            SwitchCameraMessage(type="switch_camera", rotateY=rotate_y)
        )

    def is_connected(self) -> bool:
        return self._manager.is_connected()

    def get_connection_state(self) -> ConnectionState:
        return self._manager.get_connection_state()

    async def disconnect(self) -> None:
        await self._manager.cleanup()

    def on(self, event: str, callback: Callable) -> None:
        if event == "connection_change":
            self._connection_callbacks.append(callback)
        elif event == "error":
            self._error_callbacks.append(callback)

    def off(self, event: str, callback: Callable) -> None:
        if event == "connection_change":
            try:
                self._connection_callbacks.remove(callback)
            except ValueError:
                pass
        elif event == "error":
            try:
                self._error_callbacks.remove(callback)
            except ValueError:
                pass



import logging
from typing import Optional, Callable
from dataclasses import dataclass
from aiortc import MediaStreamTrack
from tenacity import (
    retry,
    stop_after_attempt,
    wait_exponential,
    retry_if_exception,
    before_sleep_log,
)

from .webrtc_connection import WebRTCConnection
from .messages import OutgoingMessage
from .types import ConnectionState
from ..types import ModelState

logger = logging.getLogger(__name__)


@dataclass
class WebRTCConfiguration:
    webrtc_url: str
    api_key: str
    session_id: str
    fps: int
    on_remote_stream: Callable[[MediaStreamTrack], None]
    on_connection_state_change: Optional[Callable[[ConnectionState], None]] = None
    on_error: Optional[Callable[[Exception], None]] = None
    initial_state: Optional[ModelState] = None
    customize_offer: Optional[Callable] = None


def _is_retryable_error(exception: Exception) -> bool:
    """Check if an error is retryable (not a permanent error)."""
    permanent_errors = ["permission denied", "not allowed", "invalid session"]
    error_msg = str(exception).lower()
    return not any(err in error_msg for err in permanent_errors)


class WebRTCManager:
    def __init__(self, configuration: WebRTCConfiguration):
        self._config = configuration
        self._connection = self._create_connection()

    @retry(
        stop=stop_after_attempt(5),
        wait=wait_exponential(multiplier=1, min=1, max=10),
        retry=retry_if_exception(_is_retryable_error),
        before_sleep=before_sleep_log(logger, logging.WARNING),
        reraise=True,
    )
    async def connect(self, local_track: MediaStreamTrack) -> bool:
        try:
            await self._connection.connect(
                url=self._config.webrtc_url,
                local_track=local_track,
            )
            return True
        except Exception as e:
            logger.error(f"Connection attempt failed: {e}")
            await self._connection.cleanup()
            self._connection = self._create_connection()
            raise

    def _create_connection(self) -> WebRTCConnection:
        return WebRTCConnection(
            on_remote_stream=self._config.on_remote_stream,
            on_state_change=self._config.on_connection_state_change,
            on_error=self._config.on_error,
            customize_offer=self._config.customize_offer,
        )

    async def send_message(self, message: OutgoingMessage) -> None:
        await self._connection.send(message)

    async def cleanup(self) -> None:
        await self._connection.cleanup()

    def is_connected(self) -> bool:
        return self._connection.state == "connected"

    def get_connection_state(self) -> ConnectionState:
        return self._connection.state



import pytest
from decart import models, DecartSDKError


def test_realtime_models() -> None:
    model = models.realtime("mirage")
    assert model.name == "mirage"
    assert model.fps == 25
    assert model.width == 1280
    assert model.height == 704
    assert model.url_path == "/v1/stream"


def test_video_models() -> None:
    model = models.video("lucy-pro-t2v")
    assert model.name == "lucy-pro-t2v"
    assert model.url_path == "/v1/generate/lucy-pro-t2v"

    model = models.video("lucy-pro-v2v")
    assert model.name == "lucy-pro-v2v"


def test_image_models() -> None:
    model = models.image("lucy-pro-t2i")
    assert model.name == "lucy-pro-t2i"
    assert model.url_path == "/v1/generate/lucy-pro-t2i"


def test_invalid_model() -> None:
    with pytest.raises(DecartSDKError):
        models.video("invalid-model")
